from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field, model_validator


DEFAULT_LLM_MODEL = 'gpt-4o'
DEFAULT_LLM_MODEL_PROVIDER = 'openai'


class TextChunkingConfig(BaseModel):
    '''Configuration for chunking text content before they are inserted into a knowledge base'''
    separators: List[str] = Field(
        default=['\n\n', '\n', ' ', ''],
        description='List of separators to use for splitting text, in order of priority'
    )
    chunk_size: int = Field(
        default=1000,
        description='The target size of each text chunk',
        gt=0
    )
    chunk_overlap: int = Field(
        default=200,
        description='The number of characters to overlap between chunks',
        ge=0
    )


class LLMConfig(BaseModel):
    model_name: str = Field(default=DEFAULT_LLM_MODEL, description='LLM model to use for context generation')
    provider: str = Field(default=DEFAULT_LLM_MODEL_PROVIDER, description='LLM model provider to use for context generation')
    params: Dict[str, Any] = Field(default={}, description='Additional parameters to pass in when initializing the LLM')


class ContextualConfig(BaseModel):
    '''Configuration specific to contextual preprocessing'''
    llm_config: LLMConfig = Field(
        default=LLMConfig(),
        description='LLM configuration to use for context generation'
    )
    context_template: Optional[str] = Field(
        default=None,
        description='Custom template for context generation'
    )
    chunk_size: int = Field(
        default=1000,
        description='The target size of each text chunk',
        gt=0
    )
    chunk_overlap: int = Field(
        default=200,
        description='The number of characters to overlap between chunks',
        ge=0
    )


class PreprocessingConfig(BaseModel):
    '''Complete preprocessing configuration'''
    type: Literal['contextual', 'text_chunking'] = Field(
        default='text_chunking',
        description='Type of preprocessing to apply'
    )
    contextual_config: Optional[ContextualConfig] = Field(
        default=None,
        description='Configuration for contextual preprocessing'
    )
    text_chunking_config: Optional[TextChunkingConfig] = Field(
        default=None,
        description='Configuration for text chunking preprocessing'
    )

    @model_validator(mode='after')
    def validate_config_presence(self) -> 'PreprocessingConfig':
        '''Ensure the appropriate config is present for the chosen type'''
        if self.type == 'contextual' and not self.contextual_config:
            self.contextual_config = ContextualConfig()
        if self.type == 'text_chunking' and not self.text_chunking_config:
            self.text_chunking_config = TextChunkingConfig()
        return self



from openai import OpenAI
from typing import Dict, List, Optional, Union, Iterable, TYPE_CHECKING

import minds.exceptions as exc
# from minds.knowledge_bases import KnowledgeBase, KnowledgeBaseConfig

if TYPE_CHECKING:
    from minds.client import Client


class Mind:
    def __init__(
        self,
        client: 'Client',
        name: str,
        model_name: str,
        provider: str,
        # knowledge_bases=None,
        created_at: str,
        modified_at: str,
        status: str,
        datasources: Optional[List[Dict]] = [],
        parameters: Optional[Dict] = {},
        **kwargs
    ):
        self.api = client.api
        self.client = client
        
        self.name = name
        self.model_name = model_name
        self.provider = provider
        self.parameters = parameters if parameters is not None else {}
        self.created_at = created_at
        self.modified_at = modified_at
        self.datasources = datasources
        # self.knowledge_bases = knowledge_bases
        self.status = status

    def __repr__(self):
        return (f'Mind(name={self.name}, '
                f'model_name={self.model_name}, '
                f'provider={self.provider}, '
                f'created_at="{self.created_at}", '
                f'modified_at="{self.modified_at}", '
                f'parameters={self.parameters}, '
                # f'knowledge_bases={self.knowledge_bases}, '
                f'datasources={self.datasources}, '
                f'status={self.status})')

    def add_datasource(self, datasource_name: str, tables: Optional[List[str]] = None) -> None:
        """
        Add an existing Datasource to a Mind.

        :param datasource_name: name of the datasource to add.
        :param tables: list of tables to use from the datasource, optional.
        """
        response = self.api.put(
            f'/minds/{self.name}',
            data={
                'datasources': self.datasources + [{'name': datasource_name, 'tables': tables}]
            }
        )
        updated_mind = response.json()
        self.datasources = updated_mind['datasources']
        self.status = updated_mind['status']

    def remove_datasource(self, datasource_name: str) -> None:
        """
        Remove a datasource from a Mind.

        :param datasource_name: name of the datasource to remove.
        """
        response = self.api.put(
            f'/minds/{self.name}',
            data={
                'datasources': [ds for ds in (self.datasources or []) if ds['name'] != datasource_name]
            }
        )
        updated_mind = response.json()
        self.datasources = updated_mind['datasources']
        self.status = updated_mind['status']

    # def add_knowledge_base(self, knowledge_base: Union[str, KnowledgeBase, KnowledgeBaseConfig]):
    #     """
    #     Add knowledge base to mind
    #     Knowledge base can be passed as
    #      - name, str
    #      - Knowledge base object (minds.knowledge_bases.KnowledgeBase)
    #      - Knowledge base config (minds.knowledge_bases.KnowledgeBaseConfig), in this case knowledge base will be created

    #     :param knowledge_base: input knowledge base
    #     """

    #     kb_name = self.client.minds._check_knowledge_base(knowledge_base)

    #     self.api.post(
    #         f'/projects/{self.project}/minds/{self.name}/knowledge_bases',
    #         data={
    #             'name': kb_name,
    #         }
    #     )
    #     updated = self.client.minds.get(self.name)

    #     self.knowledge_bases = updated.knowledge_bases

    # def del_knowledge_base(self, knowledge_base: Union[KnowledgeBase, str]):
    #     """
    #     Delete knowledge base from mind

    #     Knowledge base can be passed as
    #      - name, str
    #      - KnowledgeBase object (minds.knowledge_bases.KnowledgeBase)

    #     :param knowledge_base: Knowledge base to delete
    #     """
    #     if isinstance(knowledge_base, KnowledgeBase):
    #         knowledge_base = knowledge_base.name
    #     elif not isinstance(knowledge_base, str):
    #         raise ValueError(f'Unknown type of knowledge base: {knowledge_base}')
    #     self.api.delete(
    #         f'/projects/{self.project}/minds/{self.name}/knowledge_bases/{knowledge_base}',
    #     )
    #     updated = self.client.minds.get(self.name)

    #     self.knowledge_bases = updated.knowledge_bases

    def completion(self, message: str, stream: bool = False) -> Union[str, Iterable[str]]:
        """
        Call mind completion

        :param message: input question
        :param stream: to enable stream mode

        :return: string if stream mode is off or iterator of strings if stream mode is on
        """
        openai_client = OpenAI(
            api_key=self.api.api_key,
            base_url=self.api.base_url
        )
        response = openai_client.chat.completions.create(
            model=self.name,
            messages=[
                {'role': 'user', 'content': message}
            ],
            stream=stream
        )
        if stream:
            return self._stream_response(response)
        else:
            return response.choices[0].message.content

    def _stream_response(self, response) -> Iterable[str]:
        for chunk in response:
            yield chunk.choices[0].delta.content


class Minds:
    def __init__(self, client: 'Client'):
        self.api = client.api
        self.client = client

    def list(self) -> List[Mind]:
        """
        Returns list of minds

        :return: iterable
        """
        data = self.api.get(f'/minds').json()
        minds_list = []
        for item in data:
            minds_list.append(Mind(self.client, **item))
        return minds_list

    def get(self, name: str) -> Mind:
        """
        Get mind by name

        :param name: name of the mind
        :return: a mind object
        """
        item = self.api.get(f'/minds/{name}').json()
        return Mind(self.client, **item)

    # def _check_knowledge_base(self, knowledge_base) -> str:
    #     if isinstance(knowledge_base, KnowledgeBase):
    #         knowledge_base = knowledge_base.name
    #     elif isinstance(knowledge_base, KnowledgeBaseConfig):
    #         # if not exists - create
    #         try:
    #             self.client.knowledge_bases.get(knowledge_base.name)
    #         except exc.ObjectNotFound:
    #             self.client.knowledge_bases.create(knowledge_base)

    #         knowledge_base = knowledge_base.name
    #     elif not isinstance(knowledge_base, str):
    #         raise ValueError(f'Unknown type of knowledge base: {knowledge_base}')
    #     return knowledge_base

    def create(
        self,
        name: str,
        model_name: Optional[str] = None,
        provider: Optional[str] = None,
        datasources: Optional[List[Dict[str, Union[str, List[str]]]]] = None,
        # knowledge_bases=None,
        parameters=None,
        replace=False,
    ) -> Mind:
        """
        Create a new mind and return it

        Datasources should be a list of dicts with keys:
        - name: str
        - tables: Optional[List[str]]

        Knowledge base can be passed as
         - name, str
         - KnowledgeBase object (minds.knowledge_bases.KnowledgeBase)
         - Knowledge base config (minds.knowledge_bases.KnowledgeBaseConfig), in this case knowledge base will be created

        :param name: name of the mind
        :param model_name: llm model name, optional
        :param provider: llm provider, optional
        :param datasources: list of datasources used by mind, optional
        :param knowledge_bases: alter list of knowledge bases used by mind, optional
        :param parameters, dict: other parameters of the mind, optional
        :param replace: if true - to remove existing mind, default is false
        :return: created mind
        """
        if replace:
            try:
                self.get(name)
                self.drop(name)
            except exc.ObjectNotFound:
                ...

        # kb_names = []
        # if knowledge_bases:
        #     for kb in knowledge_bases:
        #         kb = self._check_knowledge_base(kb)
        #         kb_names.append(kb)

        data = {
            'name': name,
            'datasources': datasources or [],
        }
        if model_name:
            data['model_name'] = model_name
        if provider:
            data['provider'] = provider
        if parameters:
            data['parameters'] = parameters
        # if kb_names:
        #     data['knowledge_bases'] = kb_names

        response = self.api.post(
            '/minds',
            data=data
        )

        return Mind(self.client, **response.json())
    
    def update(
        self,
        name: str,
        new_name: Optional[str] = None,
        model_name: Optional[str] = None,
        provider: Optional[str] = None,
        datasources: Optional[List[Dict[str, Union[str, List[str]]]]] = None,
        parameters: Optional[Dict] = None,
    ) -> Mind:
        """
        Update an existing Mind and return it

        Datasources should be a list of dicts with keys:
        - name: str
        - tables: Optional[List[str]]

        :param name: name of the mind to update
        :param new_name: new name of the mind, optional
        :param model_name: llm model name, optional
        :param provider: llm provider, optional
        :param datasources: list of datasources used by mind, optional
        :param parameters, dict: other parameters of the mind, optional
        :return: updated mind
        """
        data = {}
        if new_name:
            data['name'] = new_name
        if model_name is not None:
            data['model_name'] = model_name
        if provider is not None:
            data['provider'] = provider
        if datasources is not None:
            data['datasources'] = datasources
        if parameters is not None:
            data['parameters'] = parameters

        response = self.api.put(
            f'/minds/{name}',
            data=data
        )

        return Mind(self.client, **response.json())

    def drop(self, name: str) -> None:
       """
       Drop mind by name

       :param name: name of the mind
       """
       self.api.delete(f'/minds/{name}')



import json
import requests

import minds.exceptions as exc


def _raise_for_status(response):
    if response.status_code == 404:
        raise exc.ObjectNotFound(response.text)

    if response.status_code == 403:
        raise exc.Forbidden(response.text)

    if response.status_code == 401:
        raise exc.Unauthorized(response.text)

    if 400 <= response.status_code < 600:
        raise exc.UnknownError(f'{response.reason}: {response.text}')


class RestAPI:
    def __init__(self, api_key, base_url=None, version='v1'):
        if base_url is None:
            base_url = 'https://mdb.ai'

        base_url = base_url.rstrip('/')
        if not base_url.endswith('/api') and not base_url.endswith(f'/api/{version}'):
            base_url = base_url + '/api'
        if version and not base_url.endswith(f'/api/{version}'):
            base_url = base_url + '/' + version.lstrip('/')
        self.api_key = api_key
        self.base_url = base_url

    def _headers(self):
        return {'Authorization': 'Bearer ' + self.api_key,  'Content-Type': 'application/json',}

    def get(self, url):
        resp = requests.get(self.base_url + url, headers=self._headers())

        _raise_for_status(resp)
        return resp

    def delete(self, url, data={}):
        resp = requests.delete(
            self.base_url + url,
            headers=self._headers(),
            json=data
        )

        _raise_for_status(resp)
        return resp

    def post(self, url, data={}):
        resp = requests.post(
            self.base_url + url,
            headers=self._headers(),
            json=data,
        )

        _raise_for_status(resp)
        return resp

    def put(self, url, data={}):
        resp = requests.put(
            self.base_url + url,
            headers=self._headers(),
            json=data,
        )

        _raise_for_status(resp)
        return resp

    def patch(self, url, data={}):
        resp = requests.patch(
            self.base_url + url,
            headers=self._headers(),
            json=data,
        )

        _raise_for_status(resp)
        return resp




from minds.rest_api import RestAPI

from minds.datasources import Datasources
from minds.knowledge_bases import KnowledgeBases
from minds.minds import Minds


class Client:

    def __init__(self, api_key, base_url=None, version='v1'):

        self.api = RestAPI(api_key, base_url, version)

        self.datasources = Datasources(self)
        # self.knowledge_bases = KnowledgeBases(self)

        self.minds = Minds(self)




class ObjectNotFound(Exception):
    ...


class ObjectNotSupported(Exception):
    ...


class Forbidden(Exception):
    ...


class Unauthorized(Exception):
    ...


class UnknownError(Exception):
    ...


class MindNameInvalid(Exception):
    ...


class DatasourceNameInvalid(Exception):
    ...


from setuptools import setup, find_packages

about = {}
with open("minds/__about__.py") as fp:
    exec(fp.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as req_file:
    requirements = req_file.read().splitlines()

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=('tests*',)),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



from typing import List, Optional, TYPE_CHECKING

from pydantic import BaseModel

import minds.exceptions as exc

if TYPE_CHECKING:
    from minds.client import Client


class Datasource(BaseModel):
    """
    Existed datasource. It is returned by this SDK when datasource is queried from server
    """
    name: str
    engine: str
    description: Optional[str] = None
    connection_data: Optional[dict] = None
    created_at: str
    modified_at: str


class Datasources:
    def __init__(self, client: 'Client'):
        self.api = client.api

    def create(
        self,
        name: str,
        engine: str,
        description: Optional[str] = None,
        connection_data: Optional[dict] = None,
        replace: bool = False,
    ):
        """
        Create new datasource.

        :param name: name of datasource.
        :param engine: type of database handler, for example 'postgres', 'mysql', ...
        :param description: str, optional, description of the database. Used by mind to know what data can be got from it.
        :param connection_data: dict, optional, credentials to connect to database.
        :return: Datasource object.
        """
        if replace:
            try:
                self.get(name)
                self.drop(name)
            except exc.ObjectNotFound:
                ...

        data = {
            'name': name,
            'engine': engine,
        }
        if connection_data is not None:
            data['connection_data'] = connection_data
        if description is not None:
            data['description'] = description

        response = self.api.post(
            '/datasources',
            data=data
        )
        return Datasource(**response.json())
    
    def update(
        self,
        name: str,
        new_name: Optional[str] = None,
        description: Optional[str] = None,
        connection_data: Optional[dict] = None,
    ):
        """
        Update existing datasource.

        :param name: name of datasource to update.
        :param new_name: new name of datasource.
        :param description: str, optional, description of the database. Used by mind to know what data can be got from it.
        :param connection_data: dict, optional, credentials to connect to database.
        :return: Datasource object.
        """
        data = {}
        if new_name is not None:
            data['name'] = new_name
        if connection_data is not None:
            data['connection_data'] = connection_data
        if description is not None:
            data['description'] = description

        response = self.api.put(
            f'/datasources/{name}',
            data=data
        )
        return Datasource(**response.json())

    def list(self) -> List[Datasource]:
        """
        Returns list of datasources

        :return: iterable datasources
        """

        data = self.api.get('/datasources').json()
        ds_list = []
        for item in data:
            ds_list.append(Datasource(**item))
        return ds_list

    def get(self, name: str) -> Datasource:
        """
        Get datasource by name

        :param name: name of datasource
        :return: datasource object
        """

        data = self.api.get(f'/datasources/{name}').json()
        return Datasource(**data)

    def drop(self, name: str):
        """
        Drop datasource by name

        :param name: name of datasource
        """
        self.api.delete(f'/datasources/{name}')



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from minds.knowledge_bases.preprocessing import PreprocessingConfig
from minds.rest_api import RestAPI


class VectorStoreConfig(BaseModel):
    '''Configuration for the underlying vector store for knowledge base embeddings'''
    engine: str
    connection_data: Dict[str, Any]
    table: str = 'embeddings'


class EmbeddingConfig(BaseModel):
    '''Configuration for embeddings to use with underlying vector store for knowledge base'''
    provider: str
    model: str
    params: Optional[Dict[str, Any]] = None


class KnowledgeBaseConfig(BaseModel):
    '''Configuration for a knowledge base'''
    name: str
    description: str
    vector_store_config: Optional[VectorStoreConfig] = None
    embedding_config: Optional[EmbeddingConfig] = None
    preprocessing_config: Optional[PreprocessingConfig] = None
    # Params to apply to retrieval pipeline.
    params: Optional[Dict] = None


class KnowledgeBaseDocument(BaseModel):
    '''Represents a document that can be inserted into a knowledge base'''
    id: Union[int, str]
    content: str
    metadata: Optional[Dict[str, Any]] = {}


class KnowledgeBase:
    def __init__(self, name, api: RestAPI):
        self.name = name
        self.api = api

    def insert_from_select(self, query: str, preprocessing_config: PreprocessingConfig = None):
        '''
        Inserts select content of a connected datasource into this knowledge base

        :param query: The SQL SELECT query to use to retrieve content to be inserted
        '''
        update_request = {
            'query': query
        }
        if preprocessing_config is not None:
            update_request['preprocessing'] = preprocessing_config.model_dump()
        _ = self.api.put(f'/knowledge_bases/{self.name}', data=update_request)

    def insert_documents(self, documents: List[KnowledgeBaseDocument], preprocessing_config: PreprocessingConfig = None):
        '''
        Inserts documents directly into this knowledge base

        :param documents: The documents to insert
        '''
        update_request = {
            'rows': [d.model_dump() for d in documents]
        }
        if preprocessing_config is not None:
            update_request['preprocessing'] = preprocessing_config.model_dump()
        _ = self.api.put(f'/knowledge_bases/{self.name}', data=update_request)

    def insert_urls(self, urls: List[str], preprocessing_config: PreprocessingConfig = None):
        '''
        Crawls URLs & inserts the retrieved webpages into this knowledge base

        :param urls: Valid URLs to crawl & insert
        '''
        update_request = {
            'urls': urls
        }
        if preprocessing_config is not None:
            update_request['preprocessing'] = preprocessing_config.model_dump()
        _ = self.api.put(f'/knowledge_bases/{self.name}', data=update_request)

    def insert_files(self, files: List[str], preprocessing_config: PreprocessingConfig = None):
        '''
        Inserts files that have already been uploaded to MindsDB into this knowledge base

        :param files: Names of preuploaded files to insert
        '''
        update_request = {
            'files': files
        }
        if preprocessing_config is not None:
            update_request['preprocessing'] = preprocessing_config.model_dump()
        _ = self.api.put(f'/knowledge_bases/{self.name}', data=update_request)


class KnowledgeBases:
    def __init__(self, client):
        self.api = client.api

    def create(self, config: KnowledgeBaseConfig) -> KnowledgeBase:
        '''
        Create new knowledge base and return it

        :param config: knowledge base configuration, properties:
           - name: str, name of knowledge base
           - description: str, description of the knowledge base. Used by minds to know what data can be retrieved.
           - vector_store_config: VectorStoreConfig, configuration for embeddings vector store.
           - embedding_config: EmbeddingConfig, configuration for embeddings.
        :return: knowledge base object
        '''
        create_request = {
            'name': config.name,
            'description': config.description
        }
        if config.vector_store_config is not None:
            vector_store_data = {
                'engine': config.vector_store_config.engine,
                'connection_data': config.vector_store_config.connection_data,
                'table': config.vector_store_config.table
            }
            create_request['vector_store'] = vector_store_data
        if config.embedding_config is not None:
            embedding_data = {
                'provider': config.embedding_config.provider,
                'name': config.embedding_config.model
            }
            if config.embedding_config.params is not None:
                embedding_data.update(config.embedding_config.params)
            create_request['embedding_model'] = embedding_data
        if config.preprocessing_config is not None:
            create_request['preprocessing'] = config.preprocessing_config.model_dump()
        if config.params is not None:
            create_request['params'] = config.params

        _ = self.api.post('/knowledge_bases', data=create_request)
        return self.get(config.name)

    def list(self) -> List[KnowledgeBase]:
        '''
        Returns list of knowledge bases

        :return: iterable knowledge bases
        '''

        list_knowledge_bases_response = self.api.get('/knowledge_bases')
        knowledge_bases = list_knowledge_bases_response.json()

        all_knowledge_bases = []
        for knowledge_base in knowledge_bases:
            all_knowledge_bases.append(KnowledgeBase(knowledge_base['name'], self.api))
        return all_knowledge_bases

    def get(self, name: str) -> KnowledgeBase:
        '''
        Get knowledge base by name

        :param name: name of knowledge base
        :return: knowledge base object
        '''

        knowledge_base_response = self.api.get(f'/knowledge_bases/{name}')
        knowledge_base = knowledge_base_response.json()
        return KnowledgeBase(knowledge_base['name'], self.api)

    def drop(self, name: str, force=False):
        '''
        Drop knowledge base by name

        :param name: name of knowledge base
        :param force: if True - remove from all minds, default: False
        '''
        data = {}
        if force:
            data = {'cascade': True}

        self.api.delete(f'/knowledge_bases/{name}', data=data)



__title__ = 'minds_sdk'
__package_name__ = 'minds'
__version__ = "2.0.1"
__description__ = 'An AI-Data Mind is an LLM with the built-in power to answer data questions for Agents'
__email__ = 'hello@mindsdb.com'
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/minds_python_sdk'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__copyright__ = 'Copyright 2024-MindsDB'
